import {
  getRelayListForUser,
  NDKFilter,
  NDKKind,
  NDKRelaySet,
  NDKSubscription,
  NDKSubscriptionCacheUsage
} from '@nostr-dev-kit/ndk'
import { UserRelaysType } from 'controllers'
import { useEffect, useState } from 'react'
import { CommentEvent, ModDetails } from 'types'
import { log, LogType } from 'utils'
import { useNDKContext } from './useNDKContext'

export const useComments = (mod: ModDetails) => {
  const { ndk } = useNDKContext()
  const [commentEvents, setCommentEvents] = useState<CommentEvent[]>([])

  useEffect(() => {
    let subscription: NDKSubscription // Define the subscription variable here for cleanup

    const setupSubscription = async () => {
      // Find the mod author's relays.
      const authorReadRelays = await getRelayListForUser(mod.author, ndk)
        .then((ndkRelayList) => {
          if (ndkRelayList) return ndkRelayList[UserRelaysType.Read]
          return [] // Return an empty array if ndkRelayList is undefined
        })
        .catch((err) => {
          log(
            true,
            LogType.Error,
            `An error occurred in fetching user's (${mod.author}) ${UserRelaysType.Read}`,
            err
          )
          return [] as string[]
        })

      const filter: NDKFilter = {
        kinds: [NDKKind.Text],
        '#a': [mod.aTag]
      }

      subscription = ndk.subscribe(
        filter,
        {
          closeOnEose: false,
          cacheUsage: NDKSubscriptionCacheUsage.CACHE_FIRST
        },
        NDKRelaySet.fromRelayUrls(authorReadRelays, ndk, true)
      )

      subscription.on('event', (ndkEvent) => {
        setCommentEvents((prev) => {
          if (prev.find((e) => e.id === ndkEvent.id)) {
            return [...prev]
          }

          const commentEvent: CommentEvent = {
            kind: NDKKind.Text,
            tags: ndkEvent.tags,
            content: ndkEvent.content,
            created_at: ndkEvent.created_at!,
            pubkey: ndkEvent.pubkey,
            id: ndkEvent.id,
            sig: ndkEvent.sig!
          }

          return [commentEvent, ...prev]
        })
      })

      subscription.start()
    }

    setupSubscription()

    // Cleanup function to stop the subscription on unmount
    return () => {
      if (subscription) {
        subscription.stop()
      }
    }
  }, [mod.aTag, mod.author, ndk])

  return {
    commentEvents,
    setCommentEvents
  }
}
